#include <unistd.h>
#include <stdlib.h>

#define __USE_XOPEN
#include <sys/poll.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <features.h>		/* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>	/* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>	/* The L2 protocols */
#endif

#include <linux/if.h>
#include <string.h>
#include <netinet/in.h>
#include <asm/system.h>

#include <packet.h>
#include <cleanup.h>
#include <capdev.h>
#include <capture.h>
#include <decode.h>
#include <preproc.h>
#include <plugin.h>

proc_decode_proto decode_proto;
proc_preproc_dispatch preproc_dispatch;
proc_serial_number serial_number;
struct proto *sllproto=NULL;

/* This is our own private data */
struct lnx_priv {
	int fd;
	char *map;
	struct tpacket_req req;
	struct iovec *ring;
	size_t maplen;
	int mtu;
	int ifindex;
};

/* Bind us to an interface by index */
int linux_bindif(int fd, int ifindex)
{
	struct sockaddr_ll addr;

	/* bind the packet socket */
	memset((char *)&addr, 0, sizeof(addr));
	addr.sll_family=AF_PACKET;
	addr.sll_protocol=__constant_htons(0x03);
	addr.sll_ifindex=ifindex;
	addr.sll_hatype=0;
	addr.sll_pkttype=0;
	addr.sll_halen=0;
	if ( bind(fd, (struct sockaddr *)&addr, sizeof(addr)) ) {
		return 0;
	}

#if 0
	/* Use promiscous mode without asking */
	if ( ifindex ) {
		struct packet_mreq m;
		m.mr_ifindex=ifindex;
		m.mr_type=PACKET_MR_PROMISC;
		m.mr_alen=0;
		if ( (setsockopt(fd, SOL_PACKET,
			PACKET_ADD_MEMBERSHIP,
			(char *)&m, sizeof(m))) ) {
			return 0;
		}
	}
#endif
	return 1;
}

/* This builds a ring buffer request structure making sure
 * that we have buffers big enough so that a frame which
 * is the size of the MTU doesn't get truncated. We also
 * need to structure things with minimum memory wastage */
void linux_mkreq(struct lnx_priv *p)
{
	unsigned int pg,bs;
	unsigned int siz;
	unsigned int mult=1;

	bs=pg=getpagesize();
	siz=p->mtu+TPACKET_HDRLEN;

	if ( bs <= siz ) {
		while(bs<siz) {
			bs+=pg;
			mult++;
		}
		
		p->req.tp_block_size=bs;
		p->req.tp_frame_size=bs/mult;
		p->req.tp_block_nr=64;
		p->req.tp_frame_nr=mult*64;
	}else{
		while( (siz*(mult+1))<=pg ) {
			mult++;
		}

		p->req.tp_block_size=pg;
		p->req.tp_frame_size=pg/mult;
		p->req.tp_block_nr=64;
		p->req.tp_frame_nr=mult*64;
	}
}

/* Bind the packet socket to a specific interface,
 * we need to lookup the ifindex from the name
 * and we also need to determine the MTU */
int linux_iface(struct lnx_priv *p, char *ifn)
{
	struct ifreq ifr;

	strncpy(ifr.ifr_name, ifn, sizeof(ifr.ifr_name));
	
	/* Lookup the MTU */
	if (ioctl(p->fd, SIOCGIFMTU, &ifr) < 0) {
		goto out_err;
	}else{
		p->mtu=ifr.ifr_ifru.ifru_mtu;
	}

	/* Lookup the ifindex */
	if (ioctl(p->fd, SIOCGIFINDEX, &ifr) < 0) {
		goto out_err;
	}else{
		p->ifindex=ifr.ifr_ifru.ifru_ivalue;
	}

	/* Actually join up to the interface */
	if ( linux_bindif(p->fd, p->ifindex) ) {
		return 1;
	}else{
		printf("linux: %s: bind(): %s\n", ifn, get_err());
		return 0;
	}

out_err:
	printf("linux: %s: ioctl(): %s\n", ifn, get_err());
	return 0;
}

/* Initialise a capture process, we open the file and then
 * return our opaque private data structure to firestorm */
void *lnx_init(char *interface)
{
	struct lnx_priv *p=malloc(sizeof(struct lnx_priv));
	int i;

	if (!p) return NULL;

	if ( !sllproto && !(sllproto=decode_proto("linux")) ) {
		free(p);
		return NULL;
	}

	/* Open the packet socket */
	if ( (p->fd=socket(PF_PACKET, SOCK_DGRAM, 0))<0 ) {
		printf("linux: socket(): %s\n", get_err());
		free(p);
		return NULL;
	}

	/* Join interfaces */
	if ( !interface || !strcmp(interface, "any") ) {
		p->ifindex=0; /* zero == all */
		if ( !linux_bindif(p->fd, p->ifindex) ) {
			printf("linux: %s: bind(): %s\n",
				interface, get_err());
			close(p->fd);
			free(p);
			return NULL;
		}

		/* TODO: enumerate all interfaces for
		 * highest possible MTU */
		p->mtu=2048-TPACKET_HDRLEN;
	}else{
		if ( !linux_iface(p, interface) ) {
			close(p->fd);
			free(p);
			return NULL;
		}
	}

	/* Setup socket for RX ring buffer */
	linux_mkreq(p);
	if ( setsockopt(p->fd,SOL_PACKET,PACKET_RX_RING,
		(char *)&p->req, sizeof(p->req)) ) {
		printf("linux: setsockopt(): %s\n", get_err());
		close(p->fd);
		free(p);
		return NULL;
	}

	/* mmap() the sucker */
	p->maplen=p->req.tp_block_size * p->req.tp_block_nr;
	if ( (p->map=mmap(NULL, p->maplen,
		PROT_READ|PROT_WRITE|PROT_EXEC,
		MAP_SHARED, p->fd, 0))==MAP_FAILED ) {
		printf("linux: mmap(): %s\n", get_err());
		close(p->fd);
		free(p);
		return NULL;
	}

	/* Allocate ring buffer iovecs */
	if ( !(p->ring=malloc(p->maplen)) ) {
		printf("linux: malloc(): %s\n", get_err());
		munmap(p->map, p->maplen);
		close(p->fd);
		free(p);
		return NULL;
	}

	/* Setup our ring buffer */
	for(i=0; i<p->req.tp_frame_nr; i++) {
		p->ring[i].iov_base=(void *)(p->map+(i*p->req.tp_frame_size));
		p->ring[i].iov_len=p->req.tp_frame_size;
	}

	/* Print information */
	printf("linux: %s: mtu=%i ifindex=%i buffer=%uKB (%u frames)\n",
		interface, p->mtu, p->ifindex,
		(p->req.tp_block_size*p->req.tp_block_nr)/1024,
		p->req.tp_frame_nr);

	return p;
}

void lnx_end(void *priv)
{
	struct lnx_priv *p=priv;
	struct tpacket_stats st;
	int len=sizeof(st);
	
	if ( !p ) return;

	/* Print statistics */
	if ( !getsockopt(p->fd, SOL_PACKET,
		PACKET_STATISTICS, (char *)&st, &len)) {
		printf("linux: received %u packets, dropped %u\n",
			st.tp_packets, st.tp_drops);
	}

	munmap(p->map, p->maplen);
	close(p->fd);
	free(p);
}

void lnx_go(void *priv, struct cap_s *c)
{
	struct lnx_priv *p=priv;
	struct tpacket_hdr *h;
	struct sockaddr_ll *sll;
	struct pollfd pfd;
	struct packet pkt;
	int i;

	memset(&pkt, 0, sizeof(pkt));
	pkt.capture=c;

	for(i=0;;) {
		if ( c->state == CAP_STATE_STOP) return;

		/* If state gets set to stop during this while loop
		 * we will not see it untill the next poll... It's a
		 * corner case really.*/
		while(*(unsigned long*)p->ring[i].iov_base) {
			h=p->ring[i].iov_base;
			sll=(void *)h + TPACKET_ALIGN(sizeof(*h));

			/* build packet */
			pkt.time.tv_sec=h->tp_sec;
			pkt.time.tv_usec=h->tp_usec;
			pkt.len=h->tp_mac+h->tp_len;
			pkt.caplen=h->tp_mac+h->tp_snaplen;
			pkt.base=h;
			pkt.end=(void *)h+pkt.caplen;

			/* Decode and dispatch */
			pkt.layer[0].proto=sllproto;
			pkt.layer[0].h.raw=sll;
			pkt.layer[0].flags=0;
			pkt.layer[0].session=NULL;
			pkt.llen=1;
			pkt.rlen=0;
			pkt.present=0;

			pkt.flags=FP_CLONE|FP_LIVE|FP_PKTTYPE;
			switch ( sll->sll_pkttype ) {
				case 0:
					pkt.flags|=FP_HOST;
					break;
				case 1:
					pkt.flags|=FP_BROADCAST;
					break;
				case 2:
					pkt.flags|=FP_MULTICAST;
					break;
				case 4:
					pkt.flags|=FP_OUTGOING;
					break;
				case 3:
				default:
					pkt.flags|=FP_PROMISC;
					break;
			}

			serial_number(&pkt.serial);

			if ( sllproto->decode(&pkt) ) {
				preproc_dispatch(&pkt);
			}

			h->tp_status=0; mb();

			i=(i>=p->req.tp_frame_nr-1) ? 0 : i+1;
		}

		pfd.fd=p->fd;
		pfd.events=POLLIN|POLLERR;
		pfd.revents=0;
		poll(&pfd, 1, -1);
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_capdev_add capdev_add;

	/* validate input */
	plugin_check(in, out);

	PLUGIN_ID("capture.linux", "Linux mmap() packet socket");
	PLUGIN_VERSION(1, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");

	if ( !(capdev_add=in->import("capdev.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(decode_proto=in->import("decode.proto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_dispatch=in->import("preproc.dispatch")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(serial_number=in->import("serial_number")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !capdev_add("linux", lnx_init, lnx_end, lnx_go) ) {
		return PLUGIN_ERR_FAIL;
	}
	
	return PLUGIN_ERR_OK;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
