#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>

#include <packet.h>
#include <plugin.h>
#include <decode.h>

#define DLT_EN10MB 1

int ether_decode(struct packet *);
int ether_dprint(struct layer *, char *, int);

/*
 * These are to be registered using the "decode.add" function.
 * These structures cannot be declared const, they must also be
 * initialised with the macros provided.
 */
struct proto ether_p=init_proto_ex("ethernet", ether_decode, ether_dprint);

/* 
 * These structures tell firestorm what relation our
 * protocols have to existing protocols. They must
 * be initialised with the provided macros. Arrays
 * must be terminated with a null element.
 */
struct proto_req ether_r[]={
	proto_request("__pcap_dlt", DLT_EN10MB),
	null_request()
};

/* Print out a description of an ethernet header */
int ether_dprint(struct layer *l, char *buf, int buflen)
{
	char *b=buf;
	char *x=l->h.raw+6;
	int c;

	/* SRC */
	for(c=0; c<6; c++, x++)
		b+=sprintf(b, "%02x%c", *x&0xFF, c==5 ? ' ' : ':');

	b+=sprintf(b, "> ");

	/* DST */
	x=l->h.raw;
	for(c=0; c<6; c++, x++)
		b+=sprintf(b, "%02x%c", *x&0xFF, c==5 ? ' ' : ':');

	b+=sprintf(b, "PROTO=0x%.4x", htons(l->h.eth->proto));

	return buflen;
}

/*
 * Return codes: if we have a use for this packet, return
 * 1, else return whatever the child returns.
 */

/* This function actually does the decoding */
int ether_decode(struct packet *p)
{
	struct proto_child *pc;
	struct layer *l=&p->layer[p->llen-1];

	/* Check we aren't overflowing the array */
	if ( p->llen >= PKT_LAYERS ) return 0;
	
	/* Fill in the next layers pointer */
	if ( (p->layer[p->llen].h.raw=
		l->h.raw+sizeof(struct pkt_ethhdr))
		>= p->end ) return 0;

	/* Find a relevent child */
	for(pc=l->proto->children; pc; pc=pc->next)
	{
		if ( l->h.eth->proto == pc->id ) {
			/* Recurse to the child protocol */
			p->layer[p->llen].flags=0;
			p->layer[p->llen].session=NULL;
			p->layer[p->llen++].proto=pc->proto;
			return pc->proto->decode(p);
		}
	}

	/* Just data */
	p->layer[p->llen++].proto=NULL;
	return 0;
}

/* Plugin entry point. This function is called straight away
 * after firestorm loads us */
int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_decode_add decode_add;
	int ok=0;
	
	/* Must call the plugin_check() macro before all else,
	 * it bails us if the plugin protocol is violated */
	plugin_check(in, out);
	
	/* Tell firestorm who we are */
	PLUGIN_ID("decode.ether", "Ethernet II");
	PLUGIN_VERSION(2, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	/* Import the hooks we need from firestorm */
	if ( !(decode_add=in->import("decode.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	/* Add all our protocols in to the decode engine */
	ok+=decode_add(&ether_p, ether_r);
	
	/* report OK if at least one of our
	 * protocol handlers got in */
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

/* If we had anything to clean up, we would
 * do it here, this is called just before being
 * unloaded */
int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
